package cz.mendelu.dnaAnalyser.analyse.palindrome.detector;

import cz.mendelu.dnaAnalyser.analyse.palindrome.model.PalindromeModel;
import cz.mendelu.dnaAnalyser.sequence.NucleicBuffer;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceData;
import cz.mendelu.dnaAnalyser.sequence.nucleotide.Nucleotide;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Iterator;
import java.util.SortedSet;

/**
 * Created by xkoloma1 on 03.03.2016.
 */
public class AtatPalindromeDetector implements PalindromeDetector {

    private static final Logger LOG = LoggerFactory.getLogger(AtatPalindromeDetector.class);

    public static boolean isAtat(NucleicBuffer nucleicBuffer) {
        if (nucleicBuffer.getLength() == 0) {
            return false;
        }

        Nucleotide current, expects = nucleicBuffer.next().supplement(nucleicBuffer.getType());
        while (nucleicBuffer.hasNext()) {
            current = nucleicBuffer.next();
            if (expects != current) {
                return false;
            }
            expects = current.supplement(nucleicBuffer.getType());
        }
        return true;
    }


    private final int minSize, maxSize, maxSpacer;

    public AtatPalindromeDetector(SortedSet<Integer> size, SortedSet<Integer> spacers) {
        this.minSize = size.first();
        this.maxSize = size.last();
        this.maxSpacer = spacers.last() - (spacers.last() % 2);
    }

    @Override
    public PalindromeMatcher findPalindrome(SequenceData sequence) {
        SimplePalindromeDetector.SimplePalindromeMatcher result = new SimplePalindromeDetector.SimplePalindromeMatcher();
        if(sequence.getLength() < minSize * 2) {
            return result;  //TODO nebo radeji vyvolat vyjimku?
        }

        Iterator<Nucleotide> i = sequence.iterator();
        Nucleotide current, expects = i.next().supplement(sequence.getType());
        int start = 0, length = 0;

        while (i.hasNext()) {
            current = i.next();
            length++;
            if (expects != current) {
                // Testuji zda sekvence atat je dostatecna na vytvoreni palindromu
                if (length >= (minSize * 2)) {
                    result.add(newPalindrome(sequence, start, length));
                }
                start += length;
                length = 0;
            }
            expects = current.supplement(sequence.getType());
        }

        // Pridani palindromu na konci genomu.
        if (length >= minSize * 2) {
            result.add(newPalindrome(sequence, start, length));
        }
        return result;
    }

    @Override
    public void setAtatFilter(boolean atatFilter) {
        // Ignore
    }

    private PalindromeModel newPalindrome(SequenceData sequence, int start, int length) {
        length -= (length % 2); // Délka musí být sudé číslo!
        int size =      Math.min(length / 2, maxSize);
        int remainder = Math.max(0, length - (size * 2));
        int spacer =    Math.min(remainder, maxSpacer);
        spacer -= (spacer % 2); // Délka spaceru musí být také sudé číslo.
        return new PalindromeModel(sequence, start, size, spacer, 0);
    }

}
