package cz.mendelu.dnaAnalyser.exporter;

import com.opencsv.CSVWriter;
import cz.mendelu.dnaAnalyser.utils.analyse.service.AbstractDataSessionService;
import cz.mendelu.dnaAnalyser.utils.exception.NotFoundException;
import cz.mendelu.dnaAnalyser.analyse.g4hunter.quadruplex.QuadruplexDataSessionService;
import cz.mendelu.dnaAnalyser.analyse.rloopr.RloopDataSessionService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import lombok.extern.slf4j.Slf4j;
import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.sequence.SequenceService;
import cz.mendelu.dnaAnalyser.utils.analyse.repository.AnalyseRepository;
import cz.mendelu.dnaAnalyser.utils.analyse.data.ExporterEnums.Analysis;
import cz.mendelu.dnaAnalyser.utils.analyse.data.ExporterEnums.Output;

import java.io.IOException;
import java.io.StringWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Optional;
import java.util.UUID;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

@Service
@Slf4j
public class ExporterService {

    @Autowired
    private QuadruplexDataSessionService quadruplexDataSessionService;

    @Autowired
    private RloopDataSessionService rloopDataSessionService;

    @Autowired
    private SequenceService sequenceService;

    @Autowired
    private AnalyseRepository analyseRepository;

    public String exportData(UUID id, Analysis analysis, Output format) {

        /**
         * General method for analysis exports
         * Decides which connection and which query to use by given analysis name
         * 
         * @param id Analysis ID
         * @param analysis name of the tool for analysis (quadruples, aggregate, rloop, ...)
         * @param format format of the output (CSV, BED)
         * @return file data
         */

        String query = "SELECT * FROM " + analysis;
        AbstractDataSessionService sessionService;

        switch(analysis) {
            case AGGREGATE:
            case QUADRUPLEX:
                sessionService = quadruplexDataSessionService;
                break;
            case RLOOP:
                sessionService = rloopDataSessionService;
                break;
            default:
                // default to quadruplex for now
                sessionService = quadruplexDataSessionService;
        }

        try (Connection connection = sessionService.openConnection(id)){
            try (Statement stmt = connection.createStatement()) {
                ResultSet rs = stmt.executeQuery(query);

                switch (format) {
                    case CSV: return transforResultSetToCsv(rs);
                    case BEDGRAPH: return transformResultSetToBedGraph(rs, id, analysis);
                    default: throw new IllegalArgumentException();
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return "No-DATA";
    }

    private String transforResultSetToCsv(ResultSet rs) throws IOException, SQLException {
        StringWriter writer = new StringWriter();
        CSVWriter csvWriter = new CSVWriter(writer, '\t', '"', ' ', "\n");
        if (rs != null) {
            csvWriter.writeAll(rs, true);
        }
        return writer.toString();
    }

    private String transformResultSetToBedGraph(ResultSet rs, UUID id, Analysis analysis) throws IOException, SQLException {

        // Map for rloopr attributes, which we want in the output file
        String rlooprDescription = "\"BedGraph output of R-loop-R analysis\"";
        ArrayList<String> rlooprColumns = new ArrayList<String>();
        rlooprColumns.add("RIZGRICHNESS");
        rlooprColumns.add("RLOOPGRICHNESS");
        rlooprColumns.add("STRAND");
        rlooprColumns.add("G3");
        rlooprColumns.add("G4");
        rlooprColumns.add("GN");

        // Map for g4Hunter attributes, which we want in the output file
        String g4HunterDescription = "\"BedGraph output of G4Hunter analysis\"";
        ArrayList<String> g4HunterColumns = new ArrayList<String>();
        g4HunterColumns.add("SCORE");

        // Map for z-DNA attributes, which we want in the output file
        String zdnaDescription = "\"BedGraph output of Z-DNA analysis\"";
        ArrayList<String> zdnaColumns = new ArrayList<>();
        zdnaColumns.add("SCORE");

        String browserHeader = "browser position ";
        String result = "track type=bedGraph name=\"%s\" " +
                        "description=%s " + 
                        "visibility=2 color=0,128,0 useScore=1\n";
        String description;
        String name;
        ArrayList<String> analysisColumnMapper;

        Integer chromStart = 0, chromEnd = 0;
        String dataValue = "";

        Integer sequenceStart = null; // position of first occurence of succesfull analysis

        Integer posInGenomeStart = 0;
        Integer posInGenomeEnd = 0;

        final String format = "%s %d %d %s\n";

        String chrom = getChromName(id);
        Pattern p = Pattern.compile("[A-Z][a-z].*_(chr[1-9XY]+):([0-9].+)-([0-9].+)"); // genome browser format, e.g. Human_chr8:128748315-128753680
        Matcher match = p.matcher(chrom);

        
        while (match.find()) {
            if (match.groupCount() == 3) { //
                chrom = match.group(1);
                posInGenomeStart = Integer.parseInt(match.group(2));
                posInGenomeEnd = Integer.parseInt(match.group(3));
            }
            else { // not a match for genome browser
                chrom = chrom.replaceAll("\\s", "_"); // replace whitespaces with underslash
            }
        }

        // sort by analysis type
        switch (analysis) {

            case QUADRUPLEX:
            case AGGREGATE:
                name = "G4 Hunter";
                description = g4HunterDescription;
                analysisColumnMapper = g4HunterColumns;
                break;

            case RLOOP:
                name = "R-loop tracker";
                description = rlooprDescription;
                analysisColumnMapper = rlooprColumns;
                break;

            case ZDNA:
                name = "Z-DNA tracker";
                description = zdnaDescription;
                analysisColumnMapper = zdnaColumns;
                break;

            default:
                name = "";
                description = "";
                analysisColumnMapper = new ArrayList<String>();;
                break;
        }

        try {
            // add analysis specific header
            result = String.format(result, name, description);

            while (rs.next()) {

                try {
                    chromStart = rs.getInt("POSITION");
                    if (sequenceStart == null) {
                        sequenceStart = chromStart;
                    }
                    chromEnd = rs.getInt("POSITION") + rs.getInt("LENGTH");

                    String strand = null;
                    Integer score = 0;
                    
                    if (analysis == Analysis.RLOOP) {
                        strand = rs.getString("STRAND");

                        Integer g3 = rs.getInt("G3");
                        Integer g4 = rs.getInt("G4");
                        Integer gn = rs.getInt("GN");

                        Double rizG = rs.getDouble("RIZGRICHNESS");
                        Double rloopG = rs.getDouble("RLOOPGRICHNESS");

                        Integer scoreCalc = (int) Math.ceil(- (g3 + g4*2 + gn*3) * rizG * rloopG);
                        score = strand.equals("+") ? Math.abs(scoreCalc) : scoreCalc;
                    }
                    else if (analysis == Analysis.QUADRUPLEX || analysis == Analysis.AGGREGATE) {
                        // G4 Hunter score calculation
                        score = (int) Math.ceil(rs.getDouble("SCORE") * 100);
                    }
                    else if (analysis == Analysis.ZDNA) {
                        score = (int) Math.ceil(rs.getDouble("SCORE"));
                    }

                    dataValue = String.format(" %d", score);

                } catch (SQLException e) { // score is for g4hunter, r
                    log.error("Error fetching analysis data: {}", e);
                }    

                result += String.format(format, 
                    chrom,
                    posInGenomeStart+chromStart,
                    posInGenomeStart+chromEnd,
                    dataValue
                );

                dataValue = "";
            }
        }
        catch (NullPointerException | NotFoundException e) {
            log.error("Error finding analysis by id: {}\n {}", id, e);
        }

        // update browser header
        browserHeader += chrom + ':' + Integer.toString(posInGenomeStart) + '-' + Integer.toString(posInGenomeEnd) + "\n";
        return browserHeader + result;
    }

    private String getChromName(UUID id) {
        // gets either sequence name or ncbi_id if present (ncbi_id has higher priority)

        Sequence sequence = analyseRepository.getSequence(id);
        UUID sequenceId = sequence.getId();

        return (sequenceService.getNcbi(sequenceId) != null) ? sequenceService.getNcbi(sequenceId) : sequenceService.getName(sequenceId);
    }

}