package cz.mendelu.dnaAnalyser.sequence.inport

import cz.mendelu.dnaAnalyser.batch.BatchRepository
import cz.mendelu.dnaAnalyser.batch.BatchService
import cz.mendelu.dnaAnalyser.sequence.NCBISequenceService
import cz.mendelu.dnaAnalyser.sequence.Sequence
import cz.mendelu.dnaAnalyser.sequence.SequenceRepository
import cz.mendelu.dnaAnalyser.sequence.SequenceType
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataRepository
import cz.mendelu.dnaAnalyser.sequence.nucleotide.Nucleotide
import cz.mendelu.dnaAnalyser.user.User
import spock.lang.Ignore
import spock.lang.Specification

import java.nio.ByteBuffer
import java.util.concurrent.atomic.AtomicBoolean
import java.util.stream.Stream

class ParallelImportNucleotideCountingTest extends Specification {
    ImportSequenceService importSequenceService
    Map<UUID, Map<String, Stream<String>>> rawBufferStorageSequenceDataRepo = new HashMap<>()
    SequenceRepository sequenceRepository

    byte[] buffer
    Integer length

    def setup() {
        sequenceRepository = Stub()
        SequenceDataRepository dataRepoStub = Stub()
        importSequenceService = new ImportSequenceService()
        importSequenceService.ncbiSequenceService = Mock(NCBISequenceService)
        importSequenceService.dataProcessor = Mock(RawDataProcessor)
        BatchService batchService = new BatchService() //real batch service needed (we need to execute import}
        batchService.batchRepository = Mock(BatchRepository)
        batchService.asynchronous = false //should be blocking
        importSequenceService.batchService = batchService
        String rawData = 'A\nCC\nTTT\nGGGG\nNN\nNNN'
        buffer = rawData.bytes
        RawDataProcessor procesor = new RawDataProcessor()
        length = rawData.replace("\n", "").length()


        sequenceRepository.save(_ as Sequence) >> { Sequence sequence -> return sequence }

        dataRepoStub.saveRawBuffer(_ as UUID, _ as String, _ as byte[]) >> { UUID bufferId, String format, byte[] buffer ->
            Map<String, Stream<String>> inner = new HashMap<>()
            Stream<String> stream = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(buffer))).lines()
            inner.put(format, stream)
            rawBufferStorageSequenceDataRepo.put(bufferId, inner)
            return null
        }

        dataRepoStub.loadRawBufferToStream(_ as Sequence, _ as String) >> { Sequence sequence, String format ->
            return rawBufferStorageSequenceDataRepo.containsKey(sequence.getId()) ?
                    rawBufferStorageSequenceDataRepo.get(sequence.getId()).getOrDefault(format, null) :
                    null
        }
        dataRepoStub.save(_ as UUID, _ as Iterator<ByteBuffer>) >> {
                //we must do this because source of this iterator is stream and they are lazy
            UUID id, Iterator<ByteBuffer> buff -> while (buff.hasNext()) buff.next()
        }

        importSequenceService.sequenceDataRepository = dataRepoStub
        procesor.sequenceDataRepository = dataRepoStub
        procesor.sequenceRepository = sequenceRepository
        importSequenceService.sequenceRepository = sequenceRepository
        importSequenceService.dataProcessor = procesor
    }

    def "Test Nucleotic and length Counts  as part of import PLAIN format"() {
        given:
        def sequence = new Sequence(
                id: UUID.randomUUID(),
                name: 'ImportPlainFormat',
                type: SequenceType.DNA,
                circular: false,
                owner: new User()
        )
        _ * sequenceRepository.findById(sequence.id) >> Optional.of(sequence)

        when:
        def imported = importSequenceService.importPlainFormatFromText(sequence, buffer)
        def result = imported.getNucleicCounts()
        then:
        result.get(Nucleotide.A) == 1
        result.get(Nucleotide.C) == 2
        result.get(Nucleotide.T) == 3
        result.get(Nucleotide.G) == 4
        result.get(Nucleotide.N) == 5
        imported.length == length
    }

    def "Test Nucleotic length Counts as part of import Fasta Format"() {
        given:
        def sequence = new Sequence(
                id: UUID.randomUUID(),
                name: 'ImportPlainFormat',
                type: SequenceType.DNA,
                circular: false,
                owner: new User()
        )
        _ * sequenceRepository.findById(sequence.id) >> Optional.of(sequence)

        when:
        def imported = importSequenceService.importFASTAFormatFromText(sequence, buffer)
        def result = imported.getNucleicCounts()

        then:
        result.get(Nucleotide.A) == 1
        result.get(Nucleotide.C) == 2
        result.get(Nucleotide.T) == 3
        result.get(Nucleotide.G) == 4
        result.get(Nucleotide.N) == 5
        imported.length == length
    }

    @Ignore
    def "Test Nucleotic length Counts as part of import NCBI"() {
        given:
        def ncbiID = "MY NEW SEQUENCE"
        Sequence sequence = new Sequence(
                id: UUID.randomUUID(),
                name: 'ImportPlainFormat',
                type: SequenceType.DNA,
                circular: false,
                owner: new User()
        )

        Map<String, Stream<String>> downloadQue = new HashMap<>()
        AtomicBoolean closed = new AtomicBoolean(false)
        downloadQue.put(ncbiID, new BufferedReader(new InputStreamReader(new ByteArrayInputStream(buffer))).lines().onClose(
                closed.set(true)
        ))
        NCBISequenceService ncbiStub = Stub()
        ncbiStub.downloadFASTA(_ as String) >> { String ncbiId -> return downloadQue.getOrDefault(ncbiId, null) }
        importSequenceService.ncbiSequenceService = ncbiStub

        when:
        def imported = importSequenceService.importNCBI(sequence, ncbiID)
        def result = imported.getNucleicCounts()
        then:
        result.get(Nucleotide.A) == 1
        result.get(Nucleotide.C) == 2
        result.get(Nucleotide.T) == 3
        result.get(Nucleotide.G) == 4
        result.get(Nucleotide.N) == 5
        imported.length == length
        closed.get()
    }
}
